@@ -22,6 +22,8 @@
 https://tools.ietf.org/html/rfc6352
 """
 
+import itertools
+
 from . import collation as _mod_collation
 from . import davcommon, webdav
 
@@ -65,11 +67,15 @@ class AddressDataProperty(davcommon.SubbedProperty):
     name = "{%s}address-data" % NAMESPACE
 
     def supported_on(self, resource):
-        return resource.get_content_type() == "text/vcard"
+        try:
+            return resource.get_content_type() == "text/vcard"
+        except KeyError:
+            # Resource doesn't have a content type
+            return False
 
     async def get_value_ext(self, href, resource, el, environ, requested):
         # TODO(jelmer): Support subproperties
-        # TODO(jelmer): Don't hardcode encoding
+        # UTF-8 encoding is required by RFC 6350 (vCard format)
         el.text = b"".join(await resource.get_body()).decode("utf-8")
 
 
@@ -94,6 +100,43 @@ class AddressbookMultiGetReporter(davcommon.MultiGetReporter):
     resource_type = ADDRESSBOOK_RESOURCE_TYPE
     data_property = AddressDataProperty()
 
+    async def report(
+        self,
+        environ,
+        body,
+        resources_by_hrefs,
+        properties,
+        base_href,
+        resource,
+        depth,
+        strict,
+    ):
+        # RFC 6352 Section 8.7 (CardDAV addressbook-multiget) specifies:
+        #   "The request MUST include a Depth: 0 header on the request."
+        #
+        # This is a client requirement, and the RFC doesn't explicitly mandate
+        # that servers MUST reject requests with other Depth values. However,
+        # in strict mode, we enforce this requirement to ensure full RFC
+        # compliance and catch misbehaving clients.
+        #
+        # In non-strict mode, we accept any Depth value for compatibility with
+        # existing clients that may send non-zero Depth headers.
+        if strict and depth != "0":
+            raise webdav.BadRequestError(
+                f"{self.name} requires Depth: 0 (RFC 6352 Section 8.7), "
+                f"got Depth: {depth}"
+            )
+        return await super().report(
+            environ,
+            body,
+            resources_by_hrefs,
+            properties,
+            base_href,
+            resource,
+            depth,
+            strict,
+        )
+
 
 class Addressbook(webdav.Collection):
     resource_types = webdav.Collection.resource_types + [ADDRESSBOOK_RESOURCE_TYPE]
@@ -285,6 +328,47 @@ def apply_prop_filter(el, ab):
     return False
 
 
+def parse_filter(filter_el, cls):
+    """Parse a CardDAV filter element and build a filter object."""
+    if filter_el is None:
+        return cls
+
+    test_name = filter_el.get("test", "anyof")
+    cls.test = {"allof": all, "anyof": any}[test_name]
+
+    for prop_el in filter_el:
+        if prop_el.tag == "{urn:ietf:params:xml:ns:carddav}prop-filter":
+            parse_prop_filter(prop_el, cls)
+        else:
+            raise AssertionError(f"unknown filter tag {prop_el.tag!r}")
+
+    return cls
+
+
+def parse_prop_filter(prop_el, filter_obj):
+    """Parse a prop-filter element and add it to the filter."""
+    name = prop_el.get("name")
+    text_match = None
+    param_filters = []
+    is_not_defined = False
+
+    for subel in prop_el:
+        if subel.tag == "{urn:ietf:params:xml:ns:carddav}is-not-defined":
+            is_not_defined = True
+        elif subel.tag == "{urn:ietf:params:xml:ns:carddav}text-match":
+            text_match = {
+                "text": subel.text or "",
+                "collation": subel.get("collation", "i;unicode-casemap"),
+                "negate_condition": subel.get("negate-condition", "no") == "yes",
+                "match_type": subel.get("match-type", "contains"),
+            }
+        elif subel.tag == "{urn:ietf:params:xml:ns:carddav}param-filter":
+            # param-filter is handled by apply_param_filter() during query execution
+            pass
+
+    filter_obj.add_prop_filter(name, text_match, param_filters, is_not_defined)
+
+
 async def apply_filter(el, resource):
     """Compile a filter element into a Python function."""
     if el is None or not list(el):
@@ -330,7 +414,7 @@ class AddressbookQueryReporter(webdav.Reporter):
                     f"Unknown tag {el.tag} in report {self.name}", strict
                 )
         if requested is None:
-            # The CardDAV RFC says that behaviour mimicks that of PROPFIND,
+            # The CardDAV RFC says that behaviour mimics that of PROPFIND,
             # and the WebDAV RFC says that no body implies {DAV}allprop
             # This isn't exactly an empty body, but close enough.
             requested = ET.Element("{DAV:}allprop")
@@ -351,21 +435,32 @@ class AddressbookQueryReporter(webdav.Reporter):
         else:
             nresults = None
 
+        def filter_fn(cls):
+            return parse_filter(filter_el, cls())
+
+        def members(collection):
+            return itertools.chain(
+                collection.addressbook_query(filter_fn),
+                collection.subcollections(),
+            )
+
         i = 0
         async for href, resource in webdav.traverse_resource(
-            base_resource, base_href, depth
+            base_resource, base_href, depth, members=members
         ):
-            if not await apply_filter(filter_el, resource):
-                continue
-            if nresults is not None and i >= nresults:
-                break
-            propstat = davcommon.get_properties_with_data(
-                self.data_property,
-                href,
-                resource,
-                properties,
-                environ,
-                requested,
-            )
-            yield webdav.Status(href, "200 OK", propstat=[s async for s in propstat])
-            i += 1
+            # Ideally traverse_resource would only return the right things.
+            if getattr(resource, "content_type", None) == "text/vcard":
+                if nresults is not None and i >= nresults:
+                    break
+                propstat = davcommon.get_properties_with_data(
+                    self.data_property,
+                    href,
+                    resource,
+                    properties,
+                    environ,
+                    requested,
+                )
+                yield webdav.Status(
+                    href, "200 OK", propstat=[s async for s in propstat]
+                )
+                i += 1
